# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Code for generating classes suitable for wrapping a table for an object
returned by GetManagedObjects().
"""

import types

from ._errors import DbusClientGenerationError
from ._errors import DbusClientMissingInterfaceError
from ._errors import DbusClientMissingPropertyError


def managed_object_builder(spec):
    """
    Returns a function that builds a method interface based on 'spec'.
    This method interface is a simple one to return the values of
    properties from a table generated by a GetManagedObjects() method call
    for the object that implements the given interface.

    Usage example:

    * spec is an xml specification for an interface in the format returned
    by the Introspect() method.
    * table is the dict associated with a particular object returned by the
    GetManagedObjects() method.

    >>> builder = gmo_reader_builder(spec)
    >>> Filesystem = types.new_class("Filesystem", bases=(object,), exec_body=builder)
    >>> fs = Filesystem(table)
    >>> fs.Pool()

    :param spec: the interface specification
    :type spec: Element
    """

    try:
        interface_name = spec.attrib['name']
    except KeyError as err:  # pragma: no cover
        raise DbusClientGenerationError(
            "No name attribute found for interface.") from err

    def build_property(name):
        """
        Build a single property getter for this class.

        :param str name: the property name

        :returns: the value of the property
        :rtype: object
        """

        def dbus_func(self):
            """
            The property getter.

            :raises: DbusClientMissingPropertyError
            """
            try:
                # pylint: disable=protected-access
                return self._table[name]
            except KeyError as err:
                fmt_str = "No entry found for interface \"%s\" and property \"%s\""
                raise DbusClientMissingPropertyError(
                    fmt_str % (interface_name,
                               name), interface_name, name) from err

        return dbus_func

    def builder(namespace):
        """
        The property class's namespace.

        :param namespace: the class's namespace
        """
        for prop in spec.findall('./property'):
            try:
                name = prop.attrib['name']
            # Should not fail if introspection data is well formed.
            except KeyError as err:  # pragma: no cover
                fmt_str = ("No name attribute found for some property "
                           "belonging to interface \"%s\"")
                raise DbusClientGenerationError(
                    fmt_str % interface_name) from err

            namespace[name] = build_property(name)

        def __init__(self, table):
            """
            The initalizer for this class.

            :raises: DbusClientMissingInterfaceError
            """
            if interface_name not in table:
                fmt_str = "No data in table for interface \"%s\" found"
                raise DbusClientMissingInterfaceError(fmt_str % interface_name,
                                                      interface_name)
            # pylint: disable=protected-access
            self._table = table[interface_name]

        namespace['__init__'] = __init__

    return builder


def managed_object_class(name, spec):
    """
    Returns a class with an __init__ function which takes one
    argument, a table which is a portion of the tree returned by
    ObjectManager.GetManagedObjects(). The constructed object contains
    a set of methods for directly accessing the properties which are stored
    in the table.

    Usage example:

    * spec is an XML specification for an interface in the format returned
    by the Introspect() method.
    * table is the dict associated with a particular object returned by the
    GetManagedObjects() method.

    >>> Filesystem = managed_object_class("Filesystem", spec)
    >>> fs = Filesystem(table)
    >>> fs.Pool()

    :param str name: the name to give the auto-generated class
    :param spec: the interface specification
    :rtype: type
    """
    return types.new_class(
        name, bases=(object, ), exec_body=managed_object_builder(spec))



"""
Test generation of class for invoking dbus methods.
"""

import random
import unittest

from hypothesis import given
from hypothesis import settings
from hypothesis import HealthCheck

from dbus_client_gen import mo_query_builder
from dbus_client_gen import managed_object_class

from dbus_client_gen._errors import DbusClientMissingInterfaceError
from dbus_client_gen._errors import DbusClientMissingPropertyError
from dbus_client_gen._errors import DbusClientMissingSearchPropertiesError
from dbus_client_gen._errors import DbusClientUniqueResultError
from dbus_client_gen._errors import DbusClientUnknownSearchPropertiesError

from ._introspect import interface_strategy


class TestCase(unittest.TestCase):
    """
    Test the behavior of various auto-generated classes
    """

    @given(interface_strategy(max_children=30).map(lambda x: x.element()))
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_managed_object(self, spec):
        """
        Test that the GMO object has the correct set of methods.

        Test that there is an exception if the table is missing its interface.

        Test that there is an exception if the subtable is missing an entry.
        """
        interface_name = spec.attrib['name']
        klass = managed_object_class(interface_name, spec)
        property_names = [p.attrib['name'] for p in spec.findall("./property")]
        self.assertTrue(all(hasattr(klass, name) for name in property_names))

        with self.assertRaises(DbusClientMissingInterfaceError):
            klass({interface_name + "x": {}})

        table = {
            interface_name:
            dict((name, random.randint(0, len(property_names)))
                 for name in property_names)
        }
        table_interface = table[interface_name]
        obj = klass(table)
        self.assertTrue(
            all(
                getattr(obj, name)() == value
                for (name, value) in table_interface.items()))

        if table_interface != dict():
            remove_name = random.choice([x for x in table_interface])
            del table_interface[remove_name]
            with self.assertRaises(DbusClientMissingPropertyError):
                getattr(obj, remove_name)()

    @given(interface_strategy(max_children=30).map(lambda x: x.element()))
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_managed_object_query(self, spec):
        """
        Test that the query returns appropriate values for its query input.
        """
        query = mo_query_builder(spec)

        bad = {"bogus": None}
        with self.assertRaises(DbusClientUnknownSearchPropertiesError) \
                as context:
            query(bad)
        exception = context.exception
        self.assertEqual(frozenset(exception.specified), frozenset(bad.keys()))

        properties = [p.attrib['name'] for p in spec.findall("./property")]
        name = spec.attrib['name']
        table = {
            'junk': {
                name: dict((k, None) for k in properties)
            },
            'other': {
                "interface": dict()
            },
            'nomatch': {
                name: dict((k, 2) for k in properties)
            },
        }

        query_object = query(dict((k, None) for k in properties))
        result = list(query_object.search(table))
        if properties != []:
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0][0], "junk")

            result = query_object.require_unique_match().search(table)
            self.assertEqual(len(list(result)), 1)

            with self.assertRaises(DbusClientMissingSearchPropertiesError):
                table = {"junk": {name: dict()}}
                list(query(dict((k, None) for k in properties)).search(table))

        else:
            self.assertEqual(len(result), 2)
            self.assertEqual(
                frozenset(x[0] for x in result), frozenset(["junk",
                                                            "nomatch"]))

            with self.assertRaises(DbusClientUniqueResultError):
                query_object.require_unique_match().search(table)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Exception hierarchy for this package.
"""


class DbusClientError(Exception):
    """
    Top-level error.
    """
    pass


class DbusClientGenerationError(DbusClientError):
    """
    Exception during generation of classes.
    """
    pass


class DbusClientRuntimeError(DbusClientError):
    """
    Exception raised during execution of generated classes.
    """

    def __init__(self, message, interface_name):
        """
        Initialize with a message and an interface name.

        :param str message: the error message
        :param str interface_name: the interface name
        """
        super(DbusClientRuntimeError, self).__init__(message)
        self.interface_name = interface_name


class DbusClientMissingSearchPropertiesError(DbusClientRuntimeError):
    """
    Exception returned when searching GMO result finds expected properties
    missing.
    """

    def __init__(self, message, interface_name, query_keys, data_keys):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        :param query_keys: names of properties used in query
        :type query_keys: list of str
        :param data_keys: the keys actually available in the data
        :type data_keys: list of str
        """
        super(DbusClientMissingSearchPropertiesError, self).__init__(
            message, interface_name)
        self.query_keys = query_keys
        self.data_keys = data_keys


class DbusClientUnknownSearchPropertiesError(DbusClientRuntimeError):
    """
    Exception returned when a query is specified with a property that is
    not found in the given interface.
    """

    def __init__(self, message, interface_name, specified, allowed):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        :param specified: the specified keys
        :type specified: list of str
        :param allowed: the allowed keys
        :type allowed: list of str
        """
        super(DbusClientUnknownSearchPropertiesError, self).__init__(
            message, interface_name)
        self.specified = specified
        self.allowed = allowed


class DbusClientMissingPropertyError(
        DbusClientRuntimeError):  # pragma: no cover
    """
    Exception returned when GMO data does not contain the property name.
    """

    def __init__(self, message, interface_name, property_name):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        :param str property_name: the name of the property to look up
        """
        super(DbusClientMissingPropertyError, self).__init__(
            message, interface_name)
        self.property_name = property_name


class DbusClientMissingInterfaceError(DbusClientRuntimeError):
    """
    Exception returned when GMO data does not contain the interface name.
    """

    def __init__(self, message, interface_name):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        """
        # Note that if this is not disabled, pylint complains about
        # super-init-not-called instead.
        # pylint: disable=useless-super-delegation
        super(DbusClientMissingInterfaceError, self).__init__(
            message, interface_name)


class DbusClientSearchConditionError(DbusClientRuntimeError):
    """
    Exception raised when the search result does not match specified
    requirements.
    """

    def __init__(self, message, interface_name):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        """
        # Note that if this is not disabled, pylint complains about
        # super-init-not-called instead.
        # pylint: disable=useless-super-delegation
        super(DbusClientSearchConditionError, self).__init__(
            message, interface_name)


class DbusClientUniqueResultError(DbusClientSearchConditionError):
    """
    Exception raised when the search result does not yield a unique item.
    """

    def __init__(self, message, interface_name, props, result):
        """
        Initialize exception.

        :param str message: the error message
        :param str interface_name: the interface name
        :param dict props: the list of properties for this interface to match
        :param list result: the list of objects found via the search string
        """
        super(DbusClientUniqueResultError, self).__init__(
            message, interface_name)
        self.props = props
        self.result = result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Version information.
"""

__version__ = '0.4'
__version_info__ = tuple(int(x) for x in __version__.split('.'))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Code for generating methods suitable for identifying objects in
the data structure returned by the GetManagedObjects() method.
"""

from ._errors import DbusClientGenerationError
from ._errors import DbusClientMissingSearchPropertiesError
from ._errors import DbusClientUniqueResultError
from ._errors import DbusClientUnknownSearchPropertiesError


class GMOQuery():
    """
    Class that implements a query on the result of a D-Bus GetManagedObjects()
    call.
    """

    def __init__(self, interface_name, props):
        """
        Initialize the query with its function, which is run on a single
        entry in the GetManagedObjects result. The function is generated from
        interface_name and props.

        :param str interface_name: the particular interface
        :param dict props: properties of the interface on which to match
        """

        def filter_func(data):
            """
            Returns true if an item should be kept, false otherwise.

            :returns: true for acceptance, false for rejection
            :rtype: bool
            :raises DbusClientMissingSearchPropertiesError:
            """
            if interface_name not in data:
                return False
            sub_table = data[interface_name]

            try:
                return all(
                    sub_table[key] == value for (key, value) in props.items())
            except KeyError as err:
                fmt_str = ("Missing properties in data for some object in "
                           "interface \"%s\": %s")
                missing = ", ".join(
                    str(x) for x in
                    frozenset(props.keys()) - frozenset(sub_table.keys()))
                raise DbusClientMissingSearchPropertiesError(
                    fmt_str % (interface_name, missing), interface_name,
                    [x for x in props.keys()],
                    [x for x in sub_table.keys()]) from err

        self._interface_name = interface_name
        self._props = props
        self._filter_func = filter_func
        self._require_unique = False

    def require_unique_match(self, value=True):
        """
        If value is True or None, the search requires the result to be unique,
        i.e. there must be exactly one match.
        """
        self._require_unique = value
        return self

    def search(self, gmo_result):
        """
        Search a GetManagedObjects() result, generating any matches.

        :raises DbusClientMissingSearchPropertiesError:

        :returns: a generator of tuples of objects matched by the search
        """
        result = ((object_path, data)
                  for (object_path, data) in gmo_result.items()
                  if self._filter_func(data))

        if self._require_unique:
            list_result = [x for x in result]
            if len(list_result) != 1:
                raise DbusClientUniqueResultError(
                    "No unique match found for interface %s and properties %s, found %s"
                    % (self._interface_name, self._props, list_result),
                    self._interface_name, self._props, list_result)
            else:
                result = (x for x in list_result)

        return result


def mo_query_builder(spec):
    """
    Returns a function that builds a GMOQuery object for an interface.

    :param spec: the specification of an interface
    :type spec: Element
    :returns: a function that builds a GMOQuery object
    :rtype: keywords -> GMOQuery
    """

    try:
        interface_name = spec.attrib['name']
    except KeyError as err:  # pragma: no cover
        raise DbusClientGenerationError(
            "No name attribute found for interface.") from err

    try:
        property_names = frozenset(
            p.attrib['name'] for p in spec.findall("./property"))
    except KeyError as err:  # pragma: no cover
        fmt_str = ("No name attribute found for some property belonging to "
                   "interface \"%s\"")
        raise DbusClientGenerationError(fmt_str % interface_name) from err

    def the_func(props=None):
        """
        Takes a list of key/value pairs representing properties
        and generates a GMOQuery object which implements the requested search.

        :param props: a specification of properties to restrict values
        :type props: dict of str * object or NoneType
        :returns: an appropriately constructed GMOQuery object
        :rtype: GMOQuery

        """
        props = dict() if props is None else props

        if not frozenset(props.keys()) <= property_names:
            fmt_str = (
                "These properties in the specified query are unknown to "
                "interface \"%s\": %s")
            unknown_properties = ", ".join(
                str(x) for x in frozenset(props.keys()) - property_names)
            raise DbusClientUnknownSearchPropertiesError(
                fmt_str % (interface_name, unknown_properties), interface_name,
                [key for key in props.keys()],
                [name for name in property_names])

        return GMOQuery(interface_name, props)

    return the_func



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Top-level classes and methods.
"""

from ._errors import DbusClientError
from ._errors import DbusClientGenerationError
from ._errors import DbusClientMissingInterfaceError
from ._errors import DbusClientMissingPropertyError
from ._errors import DbusClientMissingSearchPropertiesError
from ._errors import DbusClientRuntimeError
from ._errors import DbusClientSearchConditionError
from ._errors import DbusClientUnknownSearchPropertiesError
from ._errors import DbusClientUniqueResultError

from ._managed_objects import managed_object_class
from ._managed_objects_queries import mo_query_builder
from ._managed_objects_queries import GMOQuery



"""
D-Bus introspection data strategy.
"""

# pylint: disable=fixme
# TODO: When this strategy is fully mature, port it to a separate library.
# It is correct according to the dtd, except that attributes intended to be
# D-Bus signatures supply only a valid D-Bus signature, rather than arbitrary
# CDATA.

from abc import ABC
from abc import abstractmethod

import string

import xml.etree.ElementTree as ET

from hypothesis.strategies import builds
from hypothesis.strategies import fixed_dictionaries
from hypothesis.strategies import frozensets
from hypothesis.strategies import just
from hypothesis.strategies import recursive
from hypothesis.strategies import sampled_from
from hypothesis.strategies import text

from hs_dbus_signature import dbus_signatures

_TEXT_SET = string.ascii_letters + string.digits + string.punctuation
_TEXT_STRATEGY = text(_TEXT_SET, min_size=1)


class XMLElement(ABC):
    """
    An abstract class with a method for generating an element from its data.
    """

    def __init__(self, attrs, children):
        """
        Initializer.

        :param dict attrs: a dict of attributes
        :param set children: a set of children XMLElements
        """
        self.attrs = attrs
        self.children = children

    def element(self):
        """
        Generates an XML element from the data.

        :return: XML element
        :rtype: ET.Element
        """
        elt = ET.Element(self.name(), self.attrs)
        elt.extend(child.element() for child in self.children)
        return elt

    @abstractmethod
    def name(self):
        """
        Method which returns the tag for this class.
        """
        pass


class Annotation(XMLElement):
    """
    Class representing an annotation in the introspect dtd.
    """

    def name(self):
        return "annotation"


class Arg(XMLElement):
    """
    Class representing an arg in the introspect dtd.
    """

    @classmethod
    def name(cls):
        return "arg"


class Interface(XMLElement):
    """
    Class representing an interface in the introspect dtd.
    """

    @classmethod
    def name(cls):
        return "interface"


class Method(XMLElement):
    """
    Class representing a method in the introspect dtd.
    """

    @classmethod
    def name(cls):
        return "method"


class Node(XMLElement):
    """
    Class representing a node in the introspect dtd.
    """

    def name(self):
        return "node"


class Property(XMLElement):
    """
    Class representing a property in the introspect dtd.
    """

    @classmethod
    def name(cls):
        return "property"


class Signal(XMLElement):
    """
    Class representing a signal in the introspect dtd.
    """

    @classmethod
    def name(cls):
        return "signal"


def annotation_strategy():
    """
    Build a strategy to generate data for an introspection annotation.
    """
    return builds(Annotation,
                  fixed_dictionaries({
                      'name': _TEXT_STRATEGY,
                      'value': _TEXT_STRATEGY
                  }), just(frozenset()))


def arg_strategy():
    """
    Build a strategy to generate data for an introspection arg.
    """
    return builds(Arg,
                  fixed_dictionaries({
                      'name': _TEXT_STRATEGY,
                      'type': dbus_signatures(),
                      'direction': sampled_from(["in", "out"])
                  }), frozensets(annotation_strategy()))


def interface_strategy(*, min_children=0, max_children=None):
    """
    Build a strategy to generate data for an introspection interface.

    :param min_children: the minimum number of child elements in this interface
    :type min_children: non-negative int
    :param max_children: the maximum number of child elements in this interface
    :type max_children: non-negative int or None
    """
    return builds(Interface, fixed_dictionaries({
        'name': _TEXT_STRATEGY,
    }),
                  frozensets(
                      annotation_strategy() | property_strategy()
                      | method_strategy() | signal_strategy(),
                      min_size=min_children,
                      max_size=max_children))


def method_strategy():
    """
    Build a strategy to generate data for an introspection method.
    """
    return builds(Method, fixed_dictionaries({
        'name': _TEXT_STRATEGY,
    }), frozensets(annotation_strategy() | arg_strategy()))


def _node_function(strat):
    return builds(Node, fixed_dictionaries({
        'name': _TEXT_STRATEGY
    }), frozensets(strat))


def node_strategy():
    """
    Build a strategy to generate data for an introspection node.
    """
    return recursive(interface_strategy(), _node_function)


def property_strategy():
    """
    Build a strategy to generate data for an introspection property.
    """
    return builds(Property,
                  fixed_dictionaries({
                      'name':
                      _TEXT_STRATEGY,
                      'type':
                      dbus_signatures(),
                      'access':
                      sampled_from(["read", "write", "readwrite"])
                  }), frozensets(annotation_strategy()))


def signal_arg_strategy():
    """
    Build a strategy to generate data for an introspection arg for signals.
    """
    return builds(Arg,
                  fixed_dictionaries({
                      'name': _TEXT_STRATEGY,
                      'type': dbus_signatures(),
                  }), frozensets(annotation_strategy()))


def signal_strategy():
    """
    Build a strategy to generate data for an introspection signal.
    """
    return builds(Signal, fixed_dictionaries({
        'name': _TEXT_STRATEGY
    }), frozensets(annotation_strategy() | signal_arg_strategy()))
